#include <iostream>
#include <cstdio>

#include <llvm/Support/raw_ostream.h>
#include <llvm/System/Host.h>
#include <llvm/System/Path.h>
#include <llvm/System/Process.h>
#include <llvm/ADT/OwningPtr.h>
#include <llvm/ADT/StringRef.h>
#include <llvm/Linker.h>
#include <llvm/Module.h>
#include <llvm/Function.h>
#include <llvm/Target/TargetSelect.h>
#include <llvm/Bitcode/BitstreamWriter.h>
#include <llvm/Bitcode/ReaderWriter.h>
#include <llvm/Support/MemoryBuffer.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/JIT.h>
#include <llvm/Config/config.h>
#include <llvm/Transforms/Utils/Cloning.h>
#include <llvm/LLVMContext.h>
#include <llvm/Support/FormattedStream.h>
#include <llvm/Support/raw_ostream.h>
#include <llvm/System/DynamicLibrary.h>
#include <llvm/System/Path.h>
#include <llvm/System/DynamicLibrary.h>
#include <llvm/System/Path.h>

#include <clang/Basic/FileManager.h>
#include <clang/Basic/SourceManager.h>
#include <clang/Basic/Version.h>
#include <clang/Lex/HeaderSearch.h>
#include <clang/Lex/Preprocessor.h>
#include <clang/Frontend/InitHeaderSearch.h>
#include <clang/Frontend/TextDiagnosticPrinter.h>
#include <clang/Frontend/CompileOptions.h>
#include "clang/Frontend/InitPreprocessor.h"
#include "clang/Frontend/ASTConsumers.h"
#include <clang/CodeGen/ModuleBuilder.h>
#include <clang/AST/ASTContext.h>
#include <clang/AST/ASTConsumer.h>
#include <clang/AST/Decl.h>
#include <clang/AST/DeclarationName.h>
#include <clang/AST/Stmt.h>
#include <clang/Parse/Parser.h>
#include <clang/Basic/TargetInfo.h>
#include <clang/Sema/ParseAST.h>
#include <clang/Lex/LexDiagnostic.h>
#include <clang/Lex/MacroInfo.h>

// Private CLANG headers
#include <Sema/Sema.h>

#include <iostream>
#include <stdexcept>


class ParseEnvironment {
   
public:
   
   ParseEnvironment(const clang::LangOptions& options,
                    clang::TargetInfo& target,
                    clang::Diagnostic *diag,
                    clang::FileManager *fileMgr = 0,
                    clang::SourceManager *srcMgr = 0,
                    const std::vector<std::string>* inclPath = 0,
                    clang::PPCallbacks *callbacks = 0);
   
   clang::ASTContext * getASTContext() const;
   clang::Preprocessor * getPreprocessor() const;
   clang::SourceManager * getSourceManager() const;
   
private:
   
   llvm::OwningPtr<clang::SourceManager> m_srcMgr;     // Source manager we own
   clang::SourceManager*                 m_ext_srcMgr; // Source manager we do not own.
   llvm::OwningPtr<clang::FileManager>   m_fileMgr;
   llvm::OwningPtr<clang::HeaderSearch>  m_headerInfo;
   llvm::OwningPtr<clang::Preprocessor>  m_preproc;
   llvm::OwningPtr<clang::ASTContext>    m_ast;
   
};


ParseEnvironment::ParseEnvironment(const clang::LangOptions& options,
                                   clang::TargetInfo& target,
                                   clang::Diagnostic *diag,
                                   clang::FileManager *fileMgr,
                                   clang::SourceManager *srcMgr,
                                   const std::vector<std::string>* inclPath,
                                   clang::PPCallbacks *callbacks) :
m_srcMgr(srcMgr==0 ? new clang::SourceManager : 0),
m_ext_srcMgr(srcMgr),
m_fileMgr(fileMgr ? 0 : new clang::FileManager), // create a file manager only if we have to
m_headerInfo(new clang::HeaderSearch(fileMgr ? *fileMgr : *m_fileMgr))
{
   //------------------------------------------------------------------------
   // Fill the header database
   //------------------------------------------------------------------------
   
   clang::InitHeaderSearch hiInit(*m_headerInfo);
   hiInit.AddDefaultEnvVarPaths(options);
   hiInit.AddDefaultSystemIncludePaths(options, target.getTriple());
   llvm::sys::Path clangIncl(LLVM_LIBDIR, strlen(LLVM_LIBDIR));
   clangIncl.appendComponent("clang");
   clangIncl.appendComponent(CLANG_VERSION_STRING);
   clangIncl.appendComponent("include");
   hiInit.AddPath( clangIncl.c_str(), clang::InitHeaderSearch::System,
                  true, false, false, true /*ignore sysroot*/);
   if (inclPath) {
      for (std::vector<std::string>::const_iterator iP = inclPath->begin(),
           iE = inclPath->end(); iP != iE; ++iP) {
         hiInit.AddPath(*iP, clang::InitHeaderSearch::Angled,
                        true, true, false);
      }
   }
   hiInit.Realize();
   
   //-------------------------------------------------------------------------
   // Create the preprocessor
   //-------------------------------------------------------------------------      
   m_preproc.reset(new clang::Preprocessor(*diag, options, target, *getSourceManager(), *m_headerInfo));
   m_preproc->setPPCallbacks(callbacks);
   clang::PreprocessorInitOptions ppOptions;
   InitializePreprocessor(*m_preproc, ppOptions);
   m_preproc->getBuiltinInfo().InitializeBuiltins(m_preproc->getIdentifierTable(),
                                                  m_preproc->getLangOptions().NoBuiltin);
   
   
   
   m_ast.reset(new clang::ASTContext(options,
                                     *getSourceManager(),
                                     target,
                                     m_preproc->getIdentifierTable(),
                                     m_preproc->getSelectorTable(),
                                     m_preproc->getBuiltinInfo()));
}

clang::ASTContext * ParseEnvironment::getASTContext() const
{
   return m_ast.get();
}

clang::Preprocessor * ParseEnvironment::getPreprocessor() const
{
   return m_preproc.get();
}

clang::SourceManager * ParseEnvironment::getSourceManager() const
{
   return m_ext_srcMgr ? m_ext_srcMgr : m_srcMgr.get();
}

clang::LangOptions             m_lang;
clang::TargetInfo*             m_target = 0;
clang::FileManager*            m_fileMgr = 0;
clang::DiagnosticClient*       m_diagClient = 0;
llvm::LLVMContext*             m_llvmContext = 0;

llvm::Module *compile( clang::TranslationUnitDecl* tu )
{
   if( !tu )
      return 0;
   
   clang::Diagnostic diag( m_diagClient );
   diag.setSuppressSystemWarnings( true );
   
   llvm::OwningPtr<clang::CodeGenerator> codeGen;
   clang::CompileOptions options;
   codeGen.reset(CreateLLVMCodeGen(diag, "SOME NAME [Interpreter::compile()]",
                                   options, *m_llvmContext));
   
   codeGen->Initialize(tu->getASTContext());
   
   for( clang::TranslationUnitDecl::decl_iterator it = tu->decls_begin(),
       itE = tu->decls_end(); it != itE; ++it )
      codeGen->HandleTopLevelDecl(clang::DeclGroupRef(*it));
   
   codeGen->HandleTranslationUnit(tu->getASTContext());
   
   llvm::Module* module = codeGen->ReleaseModule();
   return module;
}

ParseEnvironment* parse( const char *fileName )
{
   
   //-------------------------------------------------------------------------
   // Create diagnostics
   //-------------------------------------------------------------------------
   clang::Diagnostic diag( m_diagClient );
   diag.setSuppressSystemWarnings( true );
   
   //------------------------------------------------------------------------
   // Setup a parse environement
   //------------------------------------------------------------------------
   ParseEnvironment *pEnv = new ParseEnvironment(m_lang, *m_target, &diag, m_fileMgr, 0, 0);
   const clang::FileEntry *file = m_fileMgr->getFile( fileName );
   pEnv->getSourceManager()->createMainFileID( file, clang::SourceLocation() );
      
   llvm::raw_stdout_ostream out;
   clang::ASTConsumer* dummyConsumer = clang::CreateASTPrinter(&out);
   
   clang::Sema   sema(*pEnv->getPreprocessor(), *pEnv->getASTContext(), *dummyConsumer);
   clang::Parser p(*pEnv->getPreprocessor(), sema);
   
   pEnv->getPreprocessor()->EnterMainSourceFile();
   
   p.ParseTranslationUnit();

   if (diag.hasErrorOccurred())
      return 0;
   
   return pEnv;
}

int main(int argc, char **argv) {

   int mode = 0;
   if (argc > 1) {
      mode = atoi(argv[1]);
   }
   
   m_lang.BCPLComment = 1; // Only for C99/C++.
   m_lang.Digraphs    = 1; // C94, C99, C++.
   m_lang.CPlusPlus   = 1;
   m_lang.CXXOperatorNames = 1;
   m_lang.Bool = 1;
   
   m_lang.NeXTRuntime = 1;
   m_lang.NoInline = 1;
   
   m_lang.Exceptions = 1;
   m_lang.GNUMode = 1;
   m_lang.NoInline = 1;
   m_lang.GNUInline = 1;
   m_lang.DollarIdents = 1;
   
   m_lang.POSIXThreads = 1;   
   
   m_llvmContext = &llvm::getGlobalContext();
   m_fileMgr    = new clang::FileManager();
   
   // target:
   llvm::InitializeNativeTarget();
   m_target = clang::TargetInfo::CreateTargetInfo(llvm::sys::getHostTriple());

   m_diagClient = new clang::TextDiagnosticPrinter( llvm::errs(),
                                      true,
                                      true,
                                      true,
                                      false,
                                      false,
                                      true,
                                      llvm::sys::Process::StandardErrColumns(),
                                      true);
   m_diagClient->setLangOptions(&m_lang);
  
   if (mode == 0) {
      // JIT the declaration
      const char *fileName = "declaration.cxx";
      
      ParseEnvironment *pEnv = parse( fileName );
      clang::TranslationUnitDecl* tu = pEnv->getASTContext()->getTranslationUnitDecl();
      llvm::Module*           module = compile( tu );
      
      llvm::Linker linker( "executable", llvm::CloneModule(module) );
      module = linker.releaseModule();
      
      llvm::EngineBuilder builder(module);
      std::string errMsg;
      builder.setErrorStr(&errMsg);
      builder.setEngineKind(llvm::EngineKind::JIT);
      llvm::ExecutionEngine *engine = builder.create();
      engine->runStaticConstructorsDestructors(false);
      
      // JIT the use
      fileName = "main.cxx";
      ParseEnvironment *pEnv2 = parse( fileName );
      clang::TranslationUnitDecl* tu2 = pEnv2->getASTContext()->getTranslationUnitDecl();
      llvm::Module*           module2 = compile( tu2 );
      
      llvm::Linker linker2( "executable", llvm::CloneModule(module2) );
      module2 = linker2.releaseModule();
      engine->runStaticConstructorsDestructors(module2,false);   
      
      llvm::Function* func( module2->getFunction( "main" ) );
      std::vector<std::string> params;
      params.push_back( "executable" );
      engine->runFunctionAsMain( func,  params, 0 );
   
   }  else if (mode == 1) {
         // Link modules together.
         
         // JIT the declaration
         const char *fileName = "declaration.cxx";
         
         ParseEnvironment *pEnv = parse( fileName );
         clang::TranslationUnitDecl* tu = pEnv->getASTContext()->getTranslationUnitDecl();
         llvm::Module*           module = compile( tu );
         
         llvm::Linker linker( "executable", llvm::CloneModule(module) );
         module = linker.releaseModule();
         
         llvm::EngineBuilder builder(module);
         std::string errMsg;
         builder.setErrorStr(&errMsg);
         builder.setEngineKind(llvm::EngineKind::JIT);
         llvm::ExecutionEngine *engine = builder.create();
         engine->runStaticConstructorsDestructors(false);
         
         // JIT the use
         fileName = "main.cxx";
         ParseEnvironment *pEnv2 = parse( fileName );
         clang::TranslationUnitDecl* tu2 = pEnv2->getASTContext()->getTranslationUnitDecl();
         llvm::Module*           module2 = compile( tu2 );
         
         llvm::Linker linker2( "executable", llvm::CloneModule(module) );
         linker2.LinkInModule( llvm::CloneModule( module2 ), &errMsg );
         module2 = linker2.releaseModule();
         engine->runStaticConstructorsDestructors(module2,false);   
         
         llvm::Function* func( module2->getFunction( "main" ) );
         std::vector<std::string> params;
         params.push_back( "executable" );
         engine->runFunctionAsMain( func,  params, 0 );
         
   } else if (mode == 2) {
      // load the shared library with declaration
      std::string filename = "declaration.so";
      std::string errMsg;
      llvm::sys::Path path(filename);
      if (path.isDynamicLibrary()) {
         if (!llvm::sys::DynamicLibrary::LoadLibraryPermanently(filename.c_str(), &errMsg))
            std::cerr << "[i] Success!" << '\n';
         else {
            std::cerr << "[i] Failure: " << errMsg << '\n';
            return 1;
         }
      } else {
         return 1;
         std::cerr << "[i] Failed to load " << filename << '\n';
      }
      
      // JIT the use
      const char *fileName = "main.cxx";
      ParseEnvironment *pEnv2 = parse( fileName );
      clang::TranslationUnitDecl* tu2 = pEnv2->getASTContext()->getTranslationUnitDecl();
      llvm::Module*           module2 = compile( tu2 );
      
      llvm::Linker linker2( "executable", llvm::CloneModule(module2) );
      module2 = linker2.releaseModule();
 
      llvm::EngineBuilder builder(module2);
      builder.setErrorStr(&errMsg);
      builder.setEngineKind(llvm::EngineKind::JIT);
      llvm::ExecutionEngine *engine = builder.create();
      engine->runStaticConstructorsDestructors(false);

      llvm::Function* func( module2->getFunction( "main" ) );
      std::vector<std::string> params;
      params.push_back( "executable" );
      engine->runFunctionAsMain( func,  params, 0 );

   } else if (mode == 3) {
      // JIT the declaration
      const char *fileName = "declaration.cxx";
      
      ParseEnvironment *pEnv = parse( fileName );
      clang::TranslationUnitDecl* tu = pEnv->getASTContext()->getTranslationUnitDecl();
      llvm::Module*           module = compile( tu );
      
      llvm::Linker linker( "executable", llvm::CloneModule(module) );
      module = linker.releaseModule();
      
      llvm::EngineBuilder builder(module);
      std::string errMsg;
      builder.setErrorStr(&errMsg);
      builder.setEngineKind(llvm::EngineKind::JIT);
      llvm::ExecutionEngine *engine = builder.create();
      engine->runStaticConstructorsDestructors(false);
      
      // laod the shared library with use
      std::string filename = "main.so";
      llvm::sys::Path path(filename);
      if (path.isDynamicLibrary()) {
         std::string errMsg;
         if (!llvm::sys::DynamicLibrary::LoadLibraryPermanently(filename.c_str(), &errMsg))
            std::cerr << "[i] Success!" << std::endl;
         else {
            std::cerr << "[i] Failure: " << errMsg << std::endl;
            return 1;
         }
      } else {
         return 1;
         std::cerr << "[i] Failure: " << errMsg << std::endl;
      }
   } 
}